import java.util.Map;
import java.util.Arrays; //to print an array directly
/**
* Implementation of a game controller using a fixed state machine that is read in from a file
**/
class WuSStateMachineController implements WuSControllerInterface
{
  private WuSGameState gameState;
  private WuSInputInterface gameInput;
  
  private int movenr=1;  // in this variable we count up the move number
  private int variant=1; // in each step (move number) we can have different variants
  
  /**
  * The following map holds the complete state machine.
  * Each element maps from a condition ("movenr"|"variant"|"wolf position") to
  * a) the sheep move
  * b) the next variant
  *
  * For example: statemachine.get("1|1|.7") -> {"B1->B2","."}
  * That means: If we are in the first move, first variant and the wolf is in line 7: Move sheep B1->B2 and keep variant 1
  **/
  private Map<String,String[]> statemachine;
  
  public WuSStateMachineController(String stateMachineFileName, WuSGameState gameState, WuSInputInterface gameInput)
  {
    this.setState(gameState);
    this.setInput(gameInput);
    this.loadStateMachine(stateMachineFileName);
  }
  
  public void setInput(WuSInputInterface input)
  {
    this.gameInput = input;
  }
 public void setState(WuSGameState state)
 {
   this.gameState = state;
 }
 
 /**
 * Load the state machine from file and parse it
 **/
 private void loadStateMachine(String fileName)
{
  statemachine = new HashMap<String, String[]>();
  
  // load the file containing the state machine
  String[] lines;
  lines = loadStrings(fileName);
  
  String movenr="", variant="", wolfpos="", move="", nextvariant="";
  // now we parse the file line by line 
  for(String line: lines)
  {
    line=line.trim();
    String[] elements = line.split(" +");
    if(line.matches("^.\\..*") || line.matches("^..\\..*")) // normal line
    {
      // get all elements from that line
      movenr = elements[0].split("\\.")[0];
      variant = elements[0].split("\\.")[1];
      wolfpos = elements[2].toUpperCase();
      move = elements[3];
      nextvariant = elements[4];
    }
    else if(line.length()!=0) // continuation of last line
    {
      // we keep the movenr and variant from the last line
      wolfpos = elements[0].toUpperCase();
      move = elements[1];
      nextvariant = elements[2];
    }
    
    //finally, add the rule to our collection of rules
    statemachine.put(movenr+"|"+variant+"|"+wolfpos, new String[] {move,nextvariant});
    if(DEV_MODE)
      System.out.println(movenr+"|"+variant+"|"+wolfpos+":\t\t"+move+", \""+nextvariant+"\"");
  }

} 
 /**
 * This is where the magic happens.
 * If a new input is available, the next state is determined and the wolf and sheep positions are set accordingly.
 **/
 public void update()
 {
   if(this.gameInput.newPositionAvailable() && this.gameState.wolfLost == false) //if the player made the next move and has not lost jet
   {
     String pos = this.gameInput.getSelectedPosition();
     //first we check if this is a legal move
     //TODO
     
     String [] move=null;
     if(DEV_MODE)
       System.out.println("WuSTestController:update: New input available!: "+ this.gameInput.getSelectedPosition());
     
     //first: is there a rule for the exact position?
     if((move=statemachine.get(movenr+"|"+variant+"|"+pos))!=null)
     {
       doMove(move);
     }
     //second: is there a rule for the current row
     else if((move=statemachine.get(movenr+"|"+variant+"|."+pos.substring(1)))!=null)
     {
       doMove(move);
     }
     //third: is there a rule for all positions
     else if((move=statemachine.get(movenr+"|"+variant+"|."))!=null)
     {
       doMove(move);
     }
   }
 } //end update

/**
* This method executes a move: 
* a) It sets the sheep positions as determined in move[0] (i.e. "A1->A2")
* b) It updates the variant (if move[2] is "." the variant is not changed)
**/
 public void doMove(String [] move)
 {
   String[] sheepPositions= this.gameState.getSheepPositions();
   for(int i=0;i<sheepPositions.length;i++)
   {
     if(move[0].substring(0,2).equals(sheepPositions[i])) //find the right sheep
     {
       this.gameState.setSheepPosition(i,move[0].substring(4)); //set it to new position
     }
   }

   if(!move[1].equals(".") && !move[1].equals("!")) //"." means we dont change the variant, "!" means the wolf lost
   {
     
     this.variant=Integer.parseInt(move[1]);
   }
   if(move[1].equals("!")) //the wolf lost
   {
     this.gameState.wolfLost = true;
   }
  this.gameState.setWolfPosition(this.gameInput.getSelectedPosition());
  this.movenr++;
  if(DEV_MODE)
    System.out.println("Executing Move:  "+move[0]+"\nNew State: "+movenr+"."+this.variant + " ("+ Arrays.toString(this.gameState.getSheepPositions()) +")"); 
 } //end doMove
}
